#include "../libmpeg3.h"
#include "../mpeg3private.h"
#include "../mpeg3protos.h"
#include "mpeg3audio.h"
#include "downmix.h"

//const static float SQRT2_2 = 0.707106781; /* (SQRT(2))/2 */

// downmix stuff
static float cmixlev_lut[4] = { 0.707106781, 0.595662143, 0.500, 0.595662143 };
static float smixlev_lut[4] = { 0.707106781, 0.500, 0.0   , 0.500 };

int m_gainlevel = 400;
int m_gain2level = 0;
int m_gaincenter = 0;
int m_gainrear = 0;
int m_gainlfe = 0;

void PerformGlobalGain( short int *samples, long len )
{
	register long i;
	short left,right;
	double gainlevel, gain;


	if ( m_gainlevel<1 ) return;
	gainlevel = 1+ (m_gainlevel/100.0);

	for ( i=0; i < len * 2; i = i + 2 )
	{
			left = samples[i];
			right = samples[i+1];

			gain = (((32000-abs(left))/32000.0) * gainlevel);
			samples[i] = (short int) (left * gain);

			gain = (((32000-abs(right))/32000.0) * gainlevel);
			samples[i+1] = (short int) (right * gain);
	}
}

void PerformCurvedGlobalGain( short int *samples, long len )
{
	register long i;
	short left,right;
	double gainlevel, gain;


	if ( m_gain2level<1 ) return;
	gainlevel = 1+ (m_gain2level/100.0);

	for ( i=0; i < len * 2; i = i + 2 )
	{
			left = samples[i];
			right = samples[i+1];

			gain = (((32000-abs(left))/32000.0));
			gain = gain * gain * gainlevel;
			samples[i] = (short int) (left * gain);

			gain = (((32000-abs(right))/32000.0));
			gain = gain * gain * gainlevel;
			samples[i+1] = (short int) (right * gain);
	}
}

void PerformGlobalGain_f( float *samples, long len )
{
	register long i;
	float left,right;
	double gainlevel, gain;


	if ( m_gainlevel<1 ) return;
	gainlevel = 1+ (m_gainlevel/100.0);

	for ( i=0; i < len * 2; i = i + 2 )
	{
			left = samples[i];
			right = samples[i+1];

			gain = (((1-abs(left))/1.0) * gainlevel);
			samples[i] = (left * gain);

			gain = (((1-abs(right))/1.0) * gainlevel);
			samples[i+1] = (right * gain);
	}
}

void PerformCurvedGlobalGain_f( float *samples, long len )
{
	register long i;
	float left,right;
	double gainlevel, gain;


	if ( m_gain2level<1 ) return;
	gainlevel = 1+ (m_gain2level/100.0);

	for ( i=0; i < len * 2; i = i + 2 )
	{
			left = samples[i];
			right = samples[i+1];

			gain = (((1-abs(left))/1.0));
			gain = gain * gain * gainlevel;
			samples[i] = (left * gain);

			gain = (((1-abs(right))/1.0));
			gain = gain * gain * gainlevel;
			samples[i+1] = (right * gain);
	}
}


void PerformPreSampleGain( mpeg3audio_t *audio)
{
	register long i;
	float gainlevel, gain;
	register float *sam, *sam3, *sam4, sample;

	if ( m_gaincenter>1 ){
		gainlevel = 1+ (m_gaincenter/100.0);
		if( m_gaincenter > 999 ) gainlevel = 0;

		sam = &audio->pcm_sample[1];

		for ( i=0; i < 256; ++i )
		{
				sample = *sam;

				gain = (((1-abs(sample))/1.0));
				*sam++ = sample * gain * gain * gainlevel;
		}
	}
	if ( m_gainrear>1 ){
		gainlevel = 1+ (m_gainrear/100.0);
		if( m_gainrear > 999 ) gainlevel = 0;

		sam3 = &audio->pcm_sample[3];
		sam4 = &audio->pcm_sample[4];

		for ( i=0; i < 256; ++i )
		{
				sample = *sam3;
				gain = (((1-abs(sample))/1.0));
				*sam3++ = sample * gain * gain * gainlevel;
				sample = *sam4;
				gain = (((1-abs(sample))/1.0));
				*sam4++ = sample * gain * gain * gainlevel;
		}
	}
	if ( m_gainlfe>1 ){
		gainlevel = 1+ (m_gainlfe/100.0);
		if( m_gainlfe > 999 ) gainlevel = 0;

		sam = &audio->pcm_sample[5];

		for ( i=0; i < 256; ++i )
		{
				sample = *sam;

				gain = (((1-abs(sample))/1.0));
				*sam++ = sample * gain * gain * gainlevel;
		}
	}
}


void downmix (mpeg3audio_t *audio, 
		float *output_f, 
		short *output_i, 
		long start_position, 
		long len)
{

	PerformPreSampleGain( audio );

if(output_f)
{
	register long i, j;
	long i_end, j_end;
	dm_par_t dm_par;
						
	// compute downmix parameters
	// downmix to tow channels for now
	dm_par.clev = 0.0; dm_par.slev = 0.0; dm_par.unit = 1.0;
	if (audio->ac3_bsi.acmod & 0x1)	// have center
		dm_par.clev = cmixlev_lut[audio->ac3_bsi.cmixlev];	

	if (audio->ac3_bsi.acmod & 0x4)	// have surround channels
		dm_par.slev = smixlev_lut[audio->ac3_bsi.surmixlev];

	dm_par.unit /= 1.0 + dm_par.clev + dm_par.slev;
	dm_par.clev *= dm_par.unit;
	dm_par.slev *= dm_par.unit;

	// mix the sample, overlap
	i_end = len * 2;
	j_end = audio->pcm_size * audio->channels;
	
		switch(audio->ac3_bsi.acmod) {
		case 7:		// 3/2
			for(i = 0, j = (start_position - audio->pcm_position) * audio->channels; 
				i < i_end && j < j_end; 
				i = i+2, j += audio->channels)
			{
				output_f[i] = dm_par.unit * audio->pcm_sample[j]  + dm_par.clev * audio->pcm_sample[j+1]  + dm_par.slev * audio->pcm_sample[j+3];
				output_f[i+1] = dm_par.unit * audio->pcm_sample[j+2]  + dm_par.clev * audio->pcm_sample[j+1]  + dm_par.slev * audio->pcm_sample[j+4];
			}
			for( ; i < len * 2; i++)
			{
				output_f[i] = 0;
			}
			break;
		case 6:		// 2/2
			for(i = 0, j = (start_position - audio->pcm_position) * audio->channels; 
				i < i_end && j < j_end; 
				i = i+2, j += audio->channels)
			{
				output_f[i] = dm_par.unit * audio->pcm_sample[j]  + dm_par.slev * audio->pcm_sample[j+2];
				output_f[i+1] = dm_par.unit * audio->pcm_sample[j+1] + dm_par.slev * audio->pcm_sample[j+3];
			}
			for( ; i < len *2; i++)
			{
				output_f[i] = 0;
			}
			break;
		case 5:		// 3/1
			for(i = 0, j = (start_position - audio->pcm_position) * audio->channels; 
				i < i_end && j < j_end; 
				i = i+2, j += audio->channels)
			{
				output_f[i] = dm_par.unit * audio->pcm_sample[j]  + dm_par.clev * audio->pcm_sample[j+1]  - dm_par.slev * audio->pcm_sample[j+3];
				output_f[i+1] = dm_par.unit * audio->pcm_sample[j+2]  + dm_par.clev * audio->pcm_sample[j+1]  + dm_par.slev * audio->pcm_sample[j+3];
			}
			for( ; i < len *2; i++)
			{
				output_f[i] = 0;
			}

			break;
		case 4:		// 2/1
			for(i = 0, j = (start_position - audio->pcm_position) * audio->channels; 
				i < i_end && j < j_end; 
				i = i+2, j += audio->channels)
			{
				output_f[i] = dm_par.unit * audio->pcm_sample[j]  + dm_par.slev * audio->pcm_sample[j+2];
				output_f[i+1] = dm_par.unit * audio->pcm_sample[j+1]  + dm_par.slev * audio->pcm_sample[j+2];
			}
			for( ; i < len * 2; i++)
			{
				output_f[i] = 0;
			}
			break;
		case 3:		// 3/0
			for(i = 0, j = (start_position - audio->pcm_position) * audio->channels; 
				i < i_end && j < j_end; 
				i = i+2, j += audio->channels)
			{
				output_f[i] = dm_par.unit * audio->pcm_sample[j]  + dm_par.clev * audio->pcm_sample[j+1];
				output_f[i+1] = dm_par.unit * audio->pcm_sample[j+2]  + dm_par.clev * audio->pcm_sample[j+1];
			}
			for( ; i < len * 2; i++)
			{
				output_f[i] = 0;
			}
			break;
		case 2:		// copy to output
			for(i = 0, j = (start_position - audio->pcm_position) * audio->channels; 
				i < i_end && j < j_end; 
				i = i+2, j += audio->channels)
			{
				output_f[i] = audio->pcm_sample[j];
				output_f[i+1] = audio->pcm_sample[j+1];
			}
			for( ; i < len * 2; i++)
			{
				output_f[i] = 0;
			}
			break;
		}
		PerformGlobalGain_f( output_f, len);
		PerformCurvedGlobalGain_f( output_f, len );
}

else 
{
	register long i, j;
	long i_end, j_end;
	dm_par_t dm_par;
						
	// compute downmix parameters
	// downmix to tow channels for now
	dm_par.clev = 0.0; dm_par.slev = 0.0; dm_par.unit = 1.0;
	if (audio->ac3_bsi.acmod & 0x1)	// have center
		dm_par.clev = cmixlev_lut[audio->ac3_bsi.cmixlev];	

	if (audio->ac3_bsi.acmod & 0x4)	// have surround channels
		dm_par.slev = smixlev_lut[audio->ac3_bsi.surmixlev];

	dm_par.unit /= 1.0 + dm_par.clev + dm_par.slev;
	dm_par.clev *= dm_par.unit;
	dm_par.slev *= dm_par.unit;

	i_end = len * 2;
	j_end = audio->pcm_size * audio->channels;

	// mix the sample, overlap
		switch(audio->ac3_bsi.acmod) {
		case 7:		// 3/2
			for(i = 0, j = (start_position - audio->pcm_position) * audio->channels; 
				i < i_end && j < j_end; 
				i = i+2, j += audio->channels)
			{
				output_i[i] = (int)((dm_par.unit * audio->pcm_sample[j]  + dm_par.clev * audio->pcm_sample[j+1]  + dm_par.slev * audio->pcm_sample[j+3])* 32767);
				output_i[i+1] = (int)((dm_par.unit * audio->pcm_sample[j+2]  + dm_par.clev * audio->pcm_sample[j+1]  + dm_par.slev * audio->pcm_sample[j+4])* 32767);
			}
			for( ; i < i_end; i++)
			{
				output_i[i] = 0;
			}
			break;
		case 6:		// 2/2
			for(i = 0, j = (start_position - audio->pcm_position) * audio->channels; 
				i < i_end && j < j_end; 
				i = i+2, j += audio->channels)
			{
				output_i[i] = (int)((dm_par.unit * audio->pcm_sample[j]  + dm_par.slev * audio->pcm_sample[j+2])* 32767);
				output_i[i+1] = (int)((dm_par.unit * audio->pcm_sample[j+1] + dm_par.slev * audio->pcm_sample[j+3])* 32767);
			}
			for( ; i < i_end; i++)
			{
				output_i[i] = 0;
			}
			break;
		case 5:		// 3/1
			for(i = 0, j = (start_position - audio->pcm_position) * audio->channels; 
				i < i_end && j < j_end; 
				i = i+2, j += audio->channels)
			{
				output_i[i] = (int)((dm_par.unit * audio->pcm_sample[j]  + dm_par.clev * audio->pcm_sample[j+1]  - dm_par.slev * audio->pcm_sample[j+3])* 32767);
				output_i[i+1] = (int)((dm_par.unit * audio->pcm_sample[j+2]  + dm_par.clev * audio->pcm_sample[j+1]  + dm_par.slev * audio->pcm_sample[j+3])* 32767);
			}
			for( ; i < i_end; i++)
			{
				output_i[i] = 0;
			}

			break;
		case 4:		// 2/1
			for(i = 0, j = (start_position - audio->pcm_position) * audio->channels; 
				i < i_end && j < j_end; 
				i = i+2, j += audio->channels)
			{
				output_i[i] = (int)((dm_par.unit * audio->pcm_sample[j]  + dm_par.slev * audio->pcm_sample[j+2])* 32767);
				output_i[i+1] = (int)((dm_par.unit * audio->pcm_sample[j+1]  + dm_par.slev * audio->pcm_sample[j+2])* 32767);
			}
			for( ; i < i_end; i++)
			{
				output_i[i] = 0;
			}
			break;
		case 3:		// 3/0
			for(i = 0, j = (start_position - audio->pcm_position) * audio->channels; 
				i < i_end && j < j_end; 
				i = i+2, j += audio->channels)
			{
				output_i[i] = (int)((dm_par.unit * audio->pcm_sample[j]  + dm_par.clev * audio->pcm_sample[j+1])* 32767);
				output_i[i+1] = (int)((dm_par.unit * audio->pcm_sample[j+2]  + dm_par.clev * audio->pcm_sample[j+1])* 32767);
			}
			for( ; i < i_end; i++)
			{
				output_i[i] = 0;
			}
			break;
		case 2:		// copy to output
			for(i = 0, j = (start_position - audio->pcm_position) * audio->channels; 
				i < i_end && j < j_end; 
				i = i+2, j += audio->channels)
			{
				output_i[i] = (int)((audio->pcm_sample[j])* 32767);
				output_i[i+1] = (int)((audio->pcm_sample[j+1])* 32767);
			}
			for( ; i < i_end; i++)
			{
				output_i[i] = 0;
			}
			break;
		}
		
		PerformGlobalGain( output_i, len );
		PerformCurvedGlobalGain( output_i, len );
}

}
